#!/usr/bin/env python3
import os
from time import sleep
import glob as glob
import json
from atlas.experiments import get_logger, init_logging, require_directory, setup_file_logger
from atlas.util.iterable import group_by

WAIT_INTERVAL = 2
EVENT_DIRECTORY = "events"
DELETE_EVENTS = True
WORK_DIR = "work"


class Event(object):
    def __init__(self, event_file):
        self.event_file = event_file

        with open(self.event_file, "r") as f:
            self.event_content = json.load(f)
            self.event_type = self.event_content["event_type"]
            self.keep_event = self.event_content.get("keep_event", False)
            self.id = self.event_content.get("id", None)


def make_id_dir(id):
    return require_directory(os.path.join(WORK_DIR, id))


def search_for_events(directory):
    event_files = glob.glob(os.path.join(directory, "atlas_*.json"))
    return event_files


def stop(event, log):
    log.info("Received STOP event")
    return True


def setup_training(event, log):
    if event.id is None:
        log.info(f"Event contains no id, cannot create annotations! (File: {event.event_file})")
        return False
    annotations = event.event_content.get("annotations", None)
    if annotations is None:
        log.info(f"Event contains no annotations, cannot create annotations! (File: {event.event_file})")
        return False

    # Create a directory to write the annotation files to
    id_dir = make_id_dir(event.id)
    annotation_dir = require_directory(os.path.join(id_dir, "annotations"))
    # Group annotations by section
    annotations_by_image = group_by(annotations, lambda item: item["image"])

    for image, annotations_for_image in annotations_by_image.items():
        content_to_write = {
            "Regions": [{"name": annotation["name"], "path": annotation["path"]} for annotation in annotations_for_image],
        }
        fname = os.path.join(annotation_dir, f"test_{image}.json")
        log.info(f"Writing annotations to {fname}")
        with open(fname, "w") as f:
            json.dump(content_to_write, f)


def unknown_event(event, log):
    log.error(f"No handler for event type {event.event_type}")
    event.keep_event = True
    return False


event_handlers = {
    "stop": stop,
    "setup_training": setup_training,
}


def main():
    init_logging()
    log = get_logger()

    log.info(f"Wait interval: {WAIT_INTERVAL}")
    log.info(f"Event directory: {EVENT_DIRECTORY}")

    # Create a directory to write logs to
    log_folder = require_directory(".")
    setup_file_logger(root=log_folder, prefix="altas_controller")
    log.info("Started ATLaS controller")

    # This signal specifies when the program will stop
    stop_signal = False

    while not stop_signal:
        if not os.path.exists(EVENT_DIRECTORY):
            log.warning(f"Event directory \"{EVENT_DIRECTORY}\" does not exist!")
            sleep(WAIT_INTERVAL)
            continue

        # log.info(f"Searching directory \"{EVENT_DIRECTORY}\" for events")
        event_files = search_for_events(EVENT_DIRECTORY)
        if event_files:
            log.info(f"Found {len(event_files)} event(s)")

        for event_file in event_files:
            event = Event(event_file)
            handler = event_handlers.get(event.event_type, unknown_event)
            stop_signal = handler(event, log)

            if not event.keep_event and DELETE_EVENTS:
                log.info(f"Deleting event file \"{event_file}\"")
                if os.path.exists(event_file):
                    os.remove(event_file)

        if not stop_signal:
            sleep(WAIT_INTERVAL)

    log.info("ATLaS controller stopped")


if __name__ == "__main__":
    main()
