def make_training_parameters(**kwargs):
    import os
    import math
    from atlas.configuration import config, constants
    from atlas.testing import TestSet
    from atlas.experiments import get_slices_in_range, determine_congruent_slices
    from atlas.configuration import data_source as ds
    from atlas.data.image_provider import CHANNEL_GRAY

    brain = 20
    labels = ["bg", "cor", "wm", "__LABEL__"]
    train_slices = [(brain, __TRAIN_FROM__), (brain, __TRAIN_TO__)]
    validation_slices = get_slices_in_range(start=__VAL_FROM__, stop=__VAL_TO__, step=1, brain=brain, pattern=constants.original_path)

    congruent_slices = {key: train_slices for key in set(train_slices + validation_slices)}
    
    train_params = config.TrainParams(
        iterations=2600,
        test_interval=100,
        snapshot=100,
        class_weights="$CLASS_WEIGHTS",
        save_best_weights=False,
        input_output_mapping=ds.InputOutputMapping(
            # Multi scale input
            input=("gray2", "gray16"),
            # Labels for supervised training
            output="labels"
        ),
        num_classes=len(labels),
        display_metrics=["accuracy", "f1", ] + ["f1_{}".format(l) for l in labels if l != "bg"],
        save_batches_interval=100,
        runtime_statistics_interval=100,
        optimizer_params=config.OptimizerParams(
            learning_rate_params=config.LearningRateParams(learning_rate=0.01,
                                                           lr_policy="piecewise",
                                                           piecewise_steps=[1000, 1400, 1800, 2200],
                                                           piecewise_values=[0.01*(0.5 ** i) for i in range(1, 5)])
        ),
        network_params=config.NetworkParams(
            make_network=make_network,
            finetune=os.path.join(constants.network_dir, "self_supervised_pretrained"),
            finetune_mapping = {
                "spacing_2_initial_0_conv": "initial_0_conv",
                "spacing_2_initial_0_bn": "initial_0_bn",
                "spacing_2_initial_1_conv": "initial_1_conv",
                "spacing_2_initial_1_bn": "initial_1_bn",

                "spacing_2_block_0_0_conv": "block_0_0_conv",
                "spacing_2_block_0_0_bn": "block_0_0_bn",
                "spacing_2_block_0_1_conv": "block_0_1_conv",
                "spacing_2_block_0_1_bn": "block_0_1_bn",

                "spacing_2_block_1_0_conv": "block_1_0_conv",
                "spacing_2_block_1_0_bn": "block_1_0_bn",
                "spacing_2_block_1_1_conv": "block_1_1_conv",
                "spacing_2_block_1_1_bn": "block_1_1_bn",

                "spacing_2_block_2_0_conv": "block_2_0_conv",
                "spacing_2_block_2_0_bn": "block_2_0_bn",
                "spacing_2_block_2_1_conv": "block_2_1_conv",
                "spacing_2_block_2_1_bn": "block_2_1_bn",

                "spacing_2_block_3_0_conv": "block_3_0_conv",
                "spacing_2_block_3_0_bn": "block_3_0_bn",
                "spacing_2_block_3_1_conv": "block_3_1_conv",
                "spacing_2_block_3_1_bn": "block_3_1_bn",

                "spacing_2_block_4_0_conv": "block_4_0_conv",
                "spacing_2_block_4_0_bn": "block_4_0_bn",
                "spacing_2_block_4_1_conv": "block_4_1_conv",
                "spacing_2_block_4_1_bn": "block_4_1_bn",
                },
            ),
        iterator_params={
            config.TRAIN_SPLIT: config.IteratorParams(
                batch_size=16,
                data_sources=[ds.ImageDataSource(name="gray2", size=2025, spacing=2, channels=CHANNEL_GRAY),
                              ds.ImageDataSource(name="gray16", size=628, spacing=16, channels=CHANNEL_GRAY),
                              ds.LabelDataSource(name="labels", size=68, spacing=16, labels=labels)],
                multi_threaded_communications=11,
                image_provider_params=config.ImageProviderParams(
                    transformation_parameters=config.TransformationParams(
                            normalize_orientation=True,
                            pre_registration_base_slice=__PREREG_SLICE__,
                            brightness_augmentation=True,
                            brightness_augmentation_range=(-0.2, +0.2),
                            contrast_augmentation=True,
                            contrast_augmentation_range=(0.9, 1.1),
                            gamma_contrast_augmentation=True,
                            gamma_contrast_augmentation_range=(0.8, 1.214),
                            random_mirroring=True,
                            random_rotation=True,
                            random_rotation_range=(-math.pi / 4, +math.pi / 4)
                        ),
                    ),
                sampler_params=config.SamplerParams(
                    sampler_mode="weighted",
                    weighted_sampler_probability_weights={label: 1 / len(labels) for label in labels},
                    weighted_sampler_labels=labels,
                    area_provider_params=config.AreaProviderParams(
                        area_mode="distance",
                        distance_area_provider_radius=2500,
                        distance_area_provider_radius_spacing=2,
                        distance_area_provider_radius_per_section=True,
                        congruent_enlargement_factor=1.05,
                        labels=labels,
                        congruent_slices=congruent_slices,
                        slices=train_slices,
                        use_pre_registration_for_congruent_sections=True,
                        congruent_mode="match",
                    ),
                ),
            ),
            config.VAL_SPLIT: config.IteratorParams(
                batch_size=5,
                data_sources=[ds.ImageDataSource(name="gray2", size=8169, spacing=2, channels=CHANNEL_GRAY),
                              ds.ImageDataSource(name="gray16", size=1396, spacing=16, channels=CHANNEL_GRAY),],
                multi_threaded_communications=11,
                # Enabled to find rare bug which blocks predictions and causes timeouts
                log_batch_communication=True,
                image_provider_params=config.ImageProviderParams(
                    transformation_parameters=config.TransformationParams(
                            normalize_orientation=True,
                            pre_registration_base_slice=__PREREG_SLICE__,
                        ),
                    ),
                sampler_params=config.SamplerParams(
                    sampler_mode="grid",
                    grid_sampler_output_format=config.OutputFormat(size=836, spacing=16),
                    image_provider_params=config.ImageProviderParams(
                       transformation_parameters=config.TransformationParams(
                               normalize_orientation=True,
                               pre_registration_base_slice=__PREREG_SLICE__,
                           ),
                       ),
                    area_provider_params=config.AreaProviderParams(
                        area_mode="distance",
                        distance_area_provider_radius=2500,
                        distance_area_provider_radius_spacing=2,
                        distance_area_provider_radius_per_section=True,
                        congruent_enlargement_factor=1.05,
                        labels=labels,
                        slices=validation_slices,
                        congruent_slices=congruent_slices,
                        # congruent_slices=determine_congruent_slices(training_sections=train_slices,
                        #                                             validation_sections=validation_slices),
                        use_pre_registration_for_congruent_sections=True,
                        congruent_mode="match",
                    )
                )
            )
        }
    )

    return train_params


def make_network(**kwargs):
    from atlas.models.segmentation.unet_multiscale import UNetMultiScale

    depth = 5
    layer_width = 2
    filters = [32, 64, 64, 128, 128]
    dilation_rate = [[1, 1, 1, 1, 1], [2, 2, 2, 2, 2]]
    max_filters = 128
    initial_filters = ((16, 16), (16, 16))
    initial_filter_size = ((5, 3), (3, 3))
    initial_strides = ((4, 1), (1, 1))

    input_shape = kwargs["input_shape"]
    input_spacing = kwargs["input_spacing"]
    num_classes = kwargs["num_classes"]
    training = kwargs.get("training", None)

    model = UNetMultiScale(depth=depth,
                    block_width=layer_width,
                    filters=filters,
                    max_filters=max_filters,
                    initial_filters=initial_filters,
                    initial_filter_size=initial_filter_size,
                    initial_strides=initial_strides,
                    dilation_rates=dilation_rate,
                    deconv_mode="upsampling",
                    spacing_upsampling_mode="advanced",
                    training=training)

    model = model.make_model(input_shapes=input_shape, input_spacings=input_spacing, num_classes=num_classes)

    return model
