#!/usr/bin/env python3
import requests
from bson.objectid import ObjectId
from flask import Flask, request
from flask_cors import CORS
from flask_restful import Resource, Api
from pymongo import MongoClient

# DATABASE_HOST = "medpc018.ime.kfa-juelich.de"
DATABASE_HOST = "localhost"
DATABASE_PORT = 27017
DATABASE_NAME = "atlas_ui_db"
# UPLOAD_ROOT = "/home/cschiffer/code/atlasui_prototype/atlas_controller/events"

ATLAS_CONTROLLER_URL = "http://localhost:8000/api"
ATLAS_CONTROLLER_SETUP_TRAINING_ENDPOINT = f"{ATLAS_CONTROLLER_URL}/setup_training"

app = Flask(__name__)
CORS(app)
api = Api(app)


class MissingIdError(Exception):
    pass


class InvalidIdError(Exception):
    pass


def create_client():
    return MongoClient(DATABASE_HOST, DATABASE_PORT)


def get_projects_table(client):
    database = client.get_database(DATABASE_NAME)
    projects = database.projects
    return projects


def check_id(id_to_check):
    with create_client() as client:
        projects = get_projects_table(client)
        if projects.find_one({"_id": ObjectId(id_to_check)}):
            return True
        else:
            return False


def generate_id():
    with create_client() as client:
        projects = get_projects_table(client)
        inserted_element = projects.insert_one({})
        new_id = str(inserted_element.inserted_id)
    return new_id


def validate_id_from_data(data):
    # Check if the given id exists and is valid
    if "id" not in data:
        raise MissingIdError("Id is missing")

    id_to_check = data["id"]
    if not check_id(id_to_check):
        raise InvalidIdError(f"Id {id_to_check} is invalid. Please request a valid id.")

    return id_to_check


def find_project_for_id(id):
    with create_client() as client:
        projects = get_projects_table(client)
        project = projects.find_one({"_id": ObjectId(id)})
        return project


def get_data_for_id(id):
    with create_client() as client:
        projects = get_projects_table(client)
        return projects.find_one({"_id": ObjectId(id)})


def add_data_for_id(id, data):
    with create_client() as client:
        projects = get_projects_table(client)
        projects.update_one({"_id": ObjectId(id)}, data)


def add_annotation(id, annotation_name, annotation_path, image):
    print(f"Adding annotation: {annotation_name}")
    add_data_for_id(id, {"$push": {
        "annotations": {
            "name": annotation_name,
            "path": annotation_path,
            "image": image,
        }
    }})


def get_json_data():
    assert request.is_json
    data = request.get_json()
    return data


# def upload_data(fname, data):
#     fname = os.path.join(UPLOAD_ROOT, fname)
#     # Write data
#     with open(fname, "w") as f:
#         json.dump(data, f, indent=4)

class RequestId(Resource):
    def post(self):
        # TODO Authentication
        # Generate entry for the database
        new_id = generate_id()
        print(f"Generated new ID: {new_id}")

        return {"id": new_id}


api.add_resource(RequestId, "/request_id")


class SubmitAnnotations(Resource):
    def post(self):
        data = get_json_data()
        id = validate_id_from_data(data)
        for annotation in data["annotations"]:
            add_annotation(id=id,
                           annotation_name=annotation["annotation_name"],
                           annotation_path=annotation["annotation_path"],
                           image=annotation["image"])


api.add_resource(SubmitAnnotations, "/submit_annotations")


class SetupTraining(Resource):
    def post(self):
        # Data is received as JSON
        json_data = get_json_data()
        # Validate ID
        id = validate_id_from_data(json_data)

        # Get all fields we need
        data_for_id = get_data_for_id(id)
        annotations = json_data.get("annotations", None)
        if annotations is None:
            raise RuntimeError(f"Missing field: \"annotations\"")

        brain = json_data.get("brain", None)
        if brain is None:
            raise RuntimeError(f"Missing field: \"brain\"")

        # Create a file and upload it to the atlas controller
        event_content = {
            "id": id,
            "brain": brain,
            "annotations": annotations,
        }
        # Send to server
        response = requests.post(ATLAS_CONTROLLER_SETUP_TRAINING_ENDPOINT, json=event_content)
        return response.json()


api.add_resource(SetupTraining, "/setup_training")

if __name__ == "__main__":
    app.run(debug=True)
